/* Public domain. */

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <errno.h>
#include "cdb.h"

extern int errno;

static inline void uint32_unpack(const char s[4], uint32 * const u)
{
  uint32 result;

  result = (unsigned char) s[3];
  result <<= 8;
  result += (unsigned char) s[2];
  result <<= 8;
  result += (unsigned char) s[1];
  result <<= 8;
  result += (unsigned char) s[0];

  *u = result;
}

void cdb_free(struct cdb *c)
{
  if (c->map) {
    munmap(c->map,c->size);
    c->map = 0;
  }
}

void cdb_init(struct cdb *c,int fd)
{
  struct stat st;
  char *x;

  c->map = NULL;
  c->fd = fd;

  if (fstat(fd,&st) == 0)
    if ((size_t) st.st_size <= (size_t) 0xffffffff) {
      x = mmap(0,st.st_size,PROT_READ,MAP_SHARED,fd,0);
      if (x + 1) {
	c->size = st.st_size;
	c->map = x;
      }
    }
}

int cdb_read(struct cdb *c,char *buf,unsigned int len,uint32 pos)
{
  if (c->map) {
    if ((pos > c->size) || (c->size - pos < len)) goto FORMAT;
    memcpy(buf,c->map + pos,len);
  }
  else {
    if (lseek(c->fd, (off_t) pos, SEEK_SET) == -1) return -1;
    while (len > 0) {
      int r;
      do {
        r = read(c->fd,buf,len);
      } while ((r == -1) && (errno == EINTR));
      if (r == -1) return -1;
      if (r == 0) goto FORMAT;
      buf += r;
      len -= r;
    }
  }
  return 0;

  FORMAT:
  errno = EPROTO;
  return -1;
}

static int match(struct cdb * const c, const char *key,
		 unsigned int len, uint32 pos)
{
  char buf[32];
  unsigned int n;

  while (len > 0U) {
    n = sizeof buf;
    if (n > len) n = len;
    if (cdb_read(c,buf,n,pos) == -1) return -1;
    if (memcmp(buf,key,n)) return 0;
    pos += n;
    key += n;
    len -= n;
  }
  return 1;
}

int cdb_findnext(struct cdb * const c, const char *key, const unsigned int len)
{
   char buf[8];
   uint32 pos;
   uint32 u;
   uint32 loop = 0U;
   uint32 hslots;
   uint32 hpos;
   uint32 khash;
   uint32 kpos;
   
   u = cdb_hash(key,len);
   if (cdb_read(c, buf, 8, (u << 3) & 2047) == -1) {
      return -1;
   }
   uint32_unpack(buf + 4,&hslots);
   if (hslots == 0U) {
      return 0;
   }
   uint32_unpack(buf,&hpos);
   khash = u;
   u >>= 8;
   u %= hslots;
   u <<= 3;
   kpos = hpos + u;
   while (loop < hslots) {
      if (cdb_read(c,buf,8,kpos) == -1) {
	 return -1;
      }
      uint32_unpack(buf + 4,&pos);
      if (pos == 0U) {
	 return 0;
      }
      loop++;
      kpos += 8;
      if (kpos == hpos + (hslots << 3)) {
	 kpos = hpos;
      }
      uint32_unpack(buf,&u);
      if (u == khash) {
	 if (cdb_read(c,buf,8,pos) == -1) {
	    return -1;
	 }
	 uint32_unpack(buf,&u);
	 if (u == len) {
	   if (match(c,key,len,pos + 8) < 0) {
	      return -1;
	   } else {
	      return 1;
	   }
	 }
      }
   }
   
   return 0;
}

int cdb_find(struct cdb * const c, const char * const key, const unsigned int len)
{
   return cdb_findnext(c, key, len);
}
