#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include "lmail.h"
#include "list.h"
#include "strtools.h"
#include "hashedlist.h"
#include "varstring.h"
#include "bufferedio.h"
#include "cdb/cdb.h"
#include "lmail-smtpd.h"
#include "checkaddr.h"

static char allowed_fqdn[UCHAR_MAX];

int check_fqdn_consistency_init(void)
{
   int t = 'Z';
   
   do {
      allowed_fqdn[t] = 1;
      allowed_fqdn[t + 'a' - 'A'] = 1;
      t--;
   } while (t >= 'A');
   t = '9';
   do {
      allowed_fqdn[t] = 1;
      t--;
   } while (t >= '0');
   allowed_fqdn['-'] = allowed_fqdn['.'] = 1;
#ifdef ALLOW_UNDERSCORE_IN_FQDN
   allowed_fqdn['_'] = 1;	       /* violates RFC */
#endif
   
   return 0;
}

int check_fqdn_consistency(char *host)
{
   if (host == NULL || *host == 0) {
      D(puts("Bad check_fqdn_consisty call"));
      return -1;
   }
   do {
      if (allowed_fqdn[* ((unsigned char *) host)] == 0) {
	 return 0;
      }
      *host = tolower(*host);
   } while (*++host != 0);
   
   return 1;
}

int check_email_consistency(char *email)
{
   if (email == NULL || *email == 0) {
      D(puts("Bad check_consistencyx call"));
      return -1;
   }
   do {
      if (*++email == 0) {
	 return -1;
      }
   } while (*email != '@');
   
   return check_fqdn_consistency(++email);
}

int check_badmailfrom(ConfigFiles * const config, const char *from)
{
   const HashedList *hl;
   const Node *node_pnt;
   const char *badmailfrom_addr;   
   int subdiv;
   
   if (config == NULL || from == NULL || *from == 0) {
      D(puts("check_badmailfrom with invalid parameters"));
      return -1;
   }
   D(puts("badmailfrom check"));
   hl = &config->badmailfrom;
   subdiv = TRIVIAL_HASH(hl, from);
   node_pnt = hl->lists[subdiv].first;
   while (node_pnt != NULL) {
      badmailfrom_addr = (const char *) node_pnt->object;
      if (*badmailfrom_addr != '@') {
	 if (strcmp(badmailfrom_addr, from) == 0) {
	    D(printf("Badmailfrom match [%s]\n", badmailfrom_addr));
	    return 1;
	 }
      }
      node_pnt = node_pnt->next;      
   }
   node_pnt = hl->lists[TRIVIAL_HASH(hl, "@")].first;   
   if (node_pnt == NULL) {
      return 0;
   }
   if ((from = strchr(from, '@')) == NULL) {
      return 0;
   }
   while (node_pnt != NULL) {
      badmailfrom_addr = (const char *) node_pnt->object;
      if (strcmp(badmailfrom_addr, from) == 0) {
	 D(printf("Badmailfrom @ match [%s]\n", badmailfrom_addr));
	 return 1;
      }
      node_pnt = node_pnt->next;      
   }      
   
   return 0;
}

int check_rcpthosts(ConfigFiles * const config, const char *to)
{
   size_t to_length;   
   const HashedList *hl;
   const Node *node_pnt;
   const char *rcpthosts_host;
   int subdiv;
   
   if (config == NULL || to == NULL || *to == 0) {
      D(puts("check_rcpthosts with invalid parameters"));
      return CHECK_RCPTHOSTS_RETURN_ERROR;
   }
   D(puts("rcpthosts check"));
   if ((to = strchr(to, '@')) == NULL) {
      return CHECK_RCPTHOSTS_RETURN_NO_FQDN;
   }
   if (*++to == 0) {
      return CHECK_RCPTHOSTS_RETURN_ERROR;
   } 
   hl = &config->rcpthosts_strict;
   subdiv = TRIVIAL_HASH(hl, to);
   node_pnt = hl->lists[subdiv].first;
   while (node_pnt != NULL) {
      rcpthosts_host = (const char *) node_pnt->object;
      if (strcmp(rcpthosts_host, to) == 0) {
	 D(printf("rcpthosts match [%s]\n", rcpthosts_host));
	 return CHECK_RCPTHOSTS_RETURN_OK;
      }
      node_pnt = node_pnt->next;      
   }
   to_length = strlen(to);
   hl = &config->rcpthosts_joker;
   subdiv = REVERSE_TRIVIAL_HASH(hl, to);
   node_pnt = hl->lists[subdiv].first;
   while (node_pnt != NULL) {
      rcpthosts_host = (const char *) node_pnt->object;
      if (strcmp_reverse(to, to_length, rcpthosts_host) == 0) {
	 D(printf("rcpthosts match [%s]\n", rcpthosts_host));
	 return CHECK_RCPTHOSTS_RETURN_OK;
      }
      node_pnt = node_pnt->next;      
   }
   
   return CHECK_RCPTHOSTS_RETURN_FORBIDDEN;
}

int check_morercpthosts(ConfigFiles * const config, const char *to)
{
   if (config == NULL || to == NULL || *to == 0) {
      D(puts("check_morercpthosts with invalid parameters"));
      return CHECK_RCPTHOSTS_RETURN_ERROR;
   }
   D(puts("morercpthosts check"));
   if ((to = strchr(to, '@')) == NULL) {
      return CHECK_RCPTHOSTS_RETURN_NO_FQDN;
   }
   if (*++to == 0 || *to == '.') {
      return CHECK_RCPTHOSTS_RETURN_ERROR;
   }
   do {
      if (cdb_find(&config->morercpthosts, to, strlen(to)) > 0) {
	 D(printf("morercpthosts match [%s]\n", to));
	 return CHECK_RCPTHOSTS_RETURN_OK;
      }
   } while ((to = strchr(to + 1, '.')) != NULL);
   
   return CHECK_RCPTHOSTS_RETURN_FORBIDDEN;
}
