#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/msg.h>
#include <pthread.h>
#include <signal.h>

#define MSGLEN 32
#define MSG_NEWCNX "newcnx"
#define MSG_ACK_NEWCNX "ack_newcnx"
#define MSG_DATA "data"
#define MSG_ACK_DATA "ack_data"
#define MSG_EOF_DATA "EOF_data"
#define MSG_ACK_EOF_DATA "ack_EOF_data"
#define MSG_ABORT "abort"

struct msgbuf *msgps, *msgpr;
int msqids, msqidr, msqids2, msqidr2;

typedef struct ClientThreadArgs_ {
   long pid;
   int fromtoshmid;
} ClientThreadArgs;

void *client_thread(void *thread_args_)
{
   ClientThreadArgs *thread_args = (ClientThreadArgs *) thread_args_;
   long pid;
   char *fromto_buf = NULL;
   char *data_buf = NULL;
   struct msgbuf *msgps2, *msgpr2;
   
   if (thread_args == NULL) {
      puts("> NULL parameter in client_thread");
      return NULL;
   }
   pid = thread_args->pid;
   if (thread_args->fromtoshmid <= 0) {
      puts("fromtoshmid <= 0 ?");
   }
   printf("fromtoshmid = [%d]\n", thread_args->fromtoshmid);
   if ((fromto_buf = shmat(thread_args->fromtoshmid, NULL, 0)) == NULL) {
      perror("shmat");
      return NULL;
   }
   puts("Mmoire attache");
   shmdt(fromto_buf);
   if ((msgps2 = malloc(sizeof *msgps2 + MSGLEN)) == NULL) {
      perror("malloc");
      return NULL;
   }
   if ((msgpr2 = malloc(sizeof *msgpr2 + MSGLEN)) == NULL) {
      perror("malloc");
      return NULL;
   }   
   printf("> Thread client pret pour le processus [%ld]\n", pid);
   for (;;) {
      puts("> Attente de " MSG_DATA " ou de " MSG_EOF_DATA);
      if (msgrcv(msqids2, msgps2, MSGLEN, pid, 0) < 0) {
	 perror("> msgrcv");
      }
      printf("> Reu : [%s]\n", msgps2->mtext);
      if (strncmp(msgps2->mtext, MSG_DATA, sizeof MSG_DATA - 1) == 0) {
	 int datashmid;
	 
	 puts("> Paquet de donnes");
	 if ((datashmid = atoi(msgps2->mtext + sizeof MSG_DATA - 1)) <= 0) {
	    puts("> shmid <= 0 ?");	 
	 }
	 printf("> shmid = [%d]\n", datashmid);
	 if ((data_buf = shmat(datashmid, NULL, 0)) == NULL) {
	    perror("> shmat");
	    return NULL;
	 }
	 puts("> Mmoire attache");
	 msgpr2->mtype = pid;
	 msgpr2->mtext[MSGLEN - 1] = 0;
	 strncpy(msgpr2->mtext, MSG_ACK_DATA, MSGLEN - 1);
	 puts("> Envoi de l'acquitement " MSG_ACK_DATA);
	 if (msgsnd(msqidr2, msgpr2, MSGLEN, 0) < 0) {
	    perror("> msgsnd");
	 }
	 shmdt(data_buf);
      } else if (strncmp(msgps2->mtext, MSG_EOF_DATA, sizeof MSG_EOF_DATA - 1) == 0) {
	 int datashmid;
	 
	 puts("> Paquet de donnes EOF");
	 if ((datashmid = atoi(msgps2->mtext + sizeof MSG_EOF_DATA - 1)) < 0) {
	    puts("> shmid < 0 ?");	 
	 }
	 printf("> shmid = [%d]\n", datashmid);
	 if ((data_buf = shmat(datashmid, NULL, 0)) == NULL) {
	    perror("> shmat");
	    return NULL;
	 }
	 puts("> Mmoire attache");
	 msgpr2->mtype = pid;
	 msgpr2->mtext[MSGLEN - 1] = 0;
	 strncpy(msgpr2->mtext, MSG_ACK_EOF_DATA, MSGLEN - 1);
	 puts("> Envoi de l'acquitement " MSG_ACK_EOF_DATA);
	 if (msgsnd(msqidr2, msgpr2, MSGLEN, 0) < 0) {
	    perror("> msgsnd");
	 }
	 shmdt(data_buf);
	 break;
      } else if (strncmp(msgps2->mtext, MSG_ABORT, sizeof MSG_ABORT - 1) == 0) {
	 puts("> Abandon volontaire");
	 break;
      }
   }
   puts("> Fin du thread");
   if (msgps2 != NULL) {
      free(msgps2);
   }
   if (msgpr2 != NULL) {
      free(msgpr2);
   }
   
   return NULL;
}

void ipc_delete(void)
{
   puts("On degage les IPC");
   if ((msqids = msgget((key_t) 0x4242, 0600)) >= 0) {
      msgctl(msqids, IPC_RMID, NULL);
   }   
   if ((msqidr = msgget((key_t) 0x4243, 0600)) >= 0) {
      msgctl(msqidr, IPC_RMID, NULL);
   }      
   if ((msqids2 = msgget((key_t) 0x4244, 0600)) >= 0) {
      msgctl(msqids2, IPC_RMID, NULL);
   }         
   if ((msqidr2 = msgget((key_t) 0x4245, 0600)) >= 0) {
      msgctl(msqidr2, IPC_RMID, NULL);
   }               
}

void free_mem(void)
{
   puts("Freeing memory");
   if (msgps != NULL) {
      free(msgps);
   }
   if (msgpr != NULL) {
      free(msgpr);
   }
}

void sig(int sg)
{
   printf("Signal %d reu\n", sg);
   signal(sig, SIG_DFL);
   exit(EXIT_FAILURE);
}

int main(void)
{
   int fromtoshmid;   
   
   ipc_delete();
   atexit(ipc_delete);
   atexit(free_mem);      
   signal(SIGHUP, sig);
   signal(SIGINT, sig);   
   signal(SIGQUIT, sig);      
   signal(SIGILL, sig);         
   signal(SIGFPE, sig);            
   signal(SIGKILL, sig);               
   signal(SIGSEGV, sig);                  
   signal(SIGPIPE, sig);
   signal(SIGTERM, sig);   
   signal(SIGUSR1, sig);      
   signal(SIGUSR2, sig);         
   
   if ((msqids = msgget((key_t) 0x4242, IPC_CREAT | 0600)) < 0) {
      perror("msgget");
      return 1;
   }
   if ((msqidr = msgget((key_t) 0x4243, IPC_CREAT | 0600)) < 0) {
      perror("msgget");
      return 1;
   }
   if ((msqids2 = msgget((key_t) 0x4244, IPC_CREAT | 0600)) < 0) {
      perror("msgget");
      return 1;
   }
   if ((msqidr2 = msgget((key_t) 0x4245, IPC_CREAT | 0600)) < 0) {
      perror("msgget");
      return 1;
   }
   if ((msgps = malloc(sizeof *msgps + MSGLEN)) == NULL) {
      perror("malloc");
      return 1;
   }
   if ((msgpr = malloc(sizeof *msgpr + MSGLEN)) == NULL) {
      perror("malloc");
      return 1;
   }
   for (;;) {
      puts("Attente de " MSG_NEWCNX);
      if (msgrcv(msqids, msgps, MSGLEN, 0, 0) < 0) {
	 perror("msgrcv");
      }
      printf("Reu : [%s]\n", msgps->mtext);
      if (strncmp(msgps->mtext, MSG_ABORT, sizeof MSG_ABORT - 1) == 0) {
	 continue;
      } else if (strncmp(msgps->mtext, MSG_NEWCNX, sizeof MSG_NEWCNX - 1) != 0) {
	 puts("On a reu des choses tranges et pas " MSG_NEWCNX " ?!");
	 continue;
      }
      if ((fromtoshmid = atoi(msgps->mtext + sizeof MSG_NEWCNX - 1)) <= 0) {
	 puts("fromtoshmid <= 0 ?");
      }
      printf("fromtoshmid = [%d]\n", fromtoshmid);            
      puts("Cration du thread");
	{
	   pthread_t *client;
	   long *client_pid;

	   if ((client = malloc(sizeof *client)) == NULL) {
	      perror("malloc");	      
	   }
	   if ((client_pid = malloc(sizeof *client_pid)) == NULL) {
	      perror("malloc");
	   }
	   *client_pid = msgps->mtype;	   
	   if (pthread_create(client, NULL, client_thread, client_pid) != 0) {
	      perror("pthread_create");
	   }	   
	   if (pthread_detach(*client) != 0) {
	      perror("pthread_detach");
	   }
	}
      puts("Thread ok");
      msgpr->mtype = msgps->mtype;
      msgpr->mtext[MSGLEN - 1] = 0;
      strncpy(msgpr->mtext, MSG_ACK_NEWCNX, MSGLEN - 1);
      puts("Envoi de l'acquitement");
      if (msgsnd(msqidr, msgpr, MSGLEN, 0) < 0) {
	 perror("msgsnd");
      }
   }
   
   return 0;
}
