#include <stdlib.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include "lmail.h"

#define TEST_64 "18446744073709551615"
#define TEST_64_N 18446744073709551615ULL
#define STRSIZE_64 (sizeof TEST_64)

int has_obsolete_vprintf(void)
{
   static int result = -1;
   char buf[sizeof TEST_64];
   
   if (result >= 0) {
      return result;
   }
   snprintf(buf, sizeof buf, "%llu", TEST_64_N);
   buf[sizeof buf - 1] = 0;   
   if (strcmp(buf, TEST_64) != 0) {
      D(printf("[%s] != [%s]\n" __ buf __ TEST_64));
      return 1;
   }
   return 0;
}

struct env_table {
   const char *name;
   const int len;
};

static const struct env_table badenv_table[] = {
     { (const char *) "IFS=", 4 },
     { (const char *) "LD_", 3 },
     { (const char *) "_RLD", 5 },
     { (const char *) "SHLIB_PATH=", 11 },
     { (const char *) "LIBPATH=", 8 },
     { (const char *) "KRB_CONF", 8 },
     { (const char *) "ENV=", 4 },
     { (const char *) "BASH_ENV=", 9 },
     { NULL, 0 }
};

extern char **environ;

void cleanenv(void)
{
   char **envp = environ;
   const struct env_table *bad;
   char **cur;
   
   for (cur = envp; *cur; cur++) {
      for (bad = badenv_table; bad->name; bad++) {
	 if (strncmp(*cur, bad->name, (size_t) bad->len) == 0) {
	    char **move;
	    
	    for (move = cur; *move; move++) {
	       *move = *(move + 1);
	    }
	    cur--;
	    break;
	 }
      }
   }
}

int change_id(const char * const user, const char * const group)
{
   struct passwd *pwd;
   struct group *grp;
   
   if (group != NULL) {
      grp = getgrnam(group);
   }
   if (user != NULL && geteuid() != 0) {
      D(puts("Need supervisor priviledges to change uid"));
      return -1;
   } else if (user == NULL) {
      return 0;
   }
   if ((pwd = getpwnam(user)) == NULL) {
      D(perror("Can't find that user name"));
      return -1;
   }
   if (group != NULL && (grp = getgrnam(group)) != NULL) {
      if (setgid(grp->gr_gid) < 0) {
	 D(perror("Unable to switch to the requested group"));
	 return -1;
      }
   } else {
      if (setgid(pwd->pw_gid) < 0) {
	 D(perror("Unable to get the default user group"));
	 return -1;
      }
   }
#ifdef HAVE_INITGROUPS
   if (group == NULL && initgroups(user, pwd->pw_gid) < 0) {
      D(perror("Can't set user groups for [%s] <%s>"));
      return -1;
   }
#endif
#ifdef HAVE_SETRESUID
   if (setresuid(pwd->pw_uid, pwd->pw_uid, 0) < 0) {
#endif
      if (setuid(pwd->pw_uid) < 0) {
	 D(perror("Unable to switch to user [%s] <%s>"));
	 return -1;
      }
      if (seteuid(pwd->pw_uid) < 0) {
	 D(perror("Unable to switch to user [%s] <%s>"));
	 return -1;
      }
#ifdef HAVE_SETRESUID
   }
#endif
   cleanenv();
   
   return 0;
}
