#include <stdlib.h>
#include <stdio.h>
#include "lmail.h"
#include "list.h"
#include "hashedlist.h"

int hashed_list_init(HashedList * const hl, unsigned int subdiv)
{
   List *lists;
   
   if (hl == NULL || subdiv == 0U) {
      D(puts("hashed_list_init called with bad args"));
      return -1;
   }
   if ((lists = malloc(subdiv * (sizeof *lists))) == NULL) {
      D(perror("malloc"));
      return -1;
   }
   hl->subdiv = subdiv;
   hl->lists = lists;
   do {
      subdiv--;
      lists[subdiv].first = lists[subdiv].last = NULL;
   } while (subdiv > 0U);
   
   return 0;
}

int hashed_list_destroy(HashedList * const hl)
{
   unsigned int subdiv;
   
   if (hl == NULL || (subdiv = hl->subdiv) == 0U) {
      D(puts("Hashed list already destroyed"));
      return -1;
   }   
   do {
      subdiv--;  
      list_destroy(&hl->lists[subdiv]);
   } while (subdiv > 0U);
   free(hl->lists);
   hl->lists = NULL;
   hl->subdiv = 0U;
   
   return 0;
}

int hashed_list_push(HashedList * const hl, void * const object,
		     const unsigned int subdiv)
{
   List *list;
   Node *node;
   
   if (hl == NULL) {
      D(puts("Null hashed list"));
      return -1;
   }
   if ((node = malloc(sizeof *node)) == NULL) {
      D(puts("malloc"));
      return -1;
   }
   D(printf("Added [%s] to the hashed list\n", (char *) object));
   if (subdiv >= hl->subdiv) {
      puts("Fatal : bug would cause a buffer overflow in hashed_list_push #1");
      return -1;
   }
   list = &hl->lists[subdiv];
   node->next = NULL;
   node->object = object;
   if (list->first == NULL) {
      list->first = list->last = node;
   } else {
      list->last->next = node;
      list->last = node;
   }
   
   return 0;
}

