/* Simple but efficient memory leaks finder */
/* (C)opyleft 2000 Frank DENIS aka Jedi/Sector One <j@4u.net> */

#ifdef USE_LEAKFINDER
# include <unistd.h>
# include <stdlib.h>
# include <stdio.h>
# include <string.h>
# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>
# include <errno.h>

# define LF_LOGFILE "leakfinder.log"

static int fd = -1;
extern int errno;

static void lf_log(const char *str, const void *buf, const size_t size,
		   const char *file, const char *fun, int line)
{
   char tmp[4242];
   snprintf(tmp, sizeof tmp, "%s:%p:%lu:%s:%s:%d\n",
	    str, buf, (unsigned long) size, file, fun, line);
   if (fd < 0) {
      fd = open(LF_LOGFILE, O_CREAT | O_TRUNC | O_NDELAY | O_WRONLY, 0600);
   }
   if (fd >= 0) {
      write(fd, tmp, strlen(tmp));
      fsync(fd);
   }   
}
		   
void *lf_malloc(size_t size, 
		const char *file, const char *fun, int line)
{
   void *buf;
# ifdef USE_LEAKFINDER_NOMEM   
   static long salt;
   
   if (salt == 0L) {
      salt = (long) getpid();
      srand(salt);
   }
   if ((rand() & 3) != 0) {
      buf = malloc(size);
   } else {
      buf = NULL;
      errno = ENOMEM;
   }
# else
   buf = malloc(size);
# endif
   if (buf != NULL) {
      lf_log("M", buf, size, file, fun, line);
   }

   return buf;
}

void *lf_realloc(void *buf, size_t size, 
		 const char *file, const char *fun, int line)
{
   if (buf != NULL) {
      lf_log("F", buf, 0U, file, fun, line);
   }
   buf = realloc(buf, size);
   lf_log("R", buf, size, file, fun, line);      

   return buf;
}

void *lf_calloc(size_t nmemb, size_t size, 
		const char *file, const char *fun, int line)
{
   void *buf = calloc(nmemb, size);

   lf_log("C", buf, nmemb * size, file, fun, line);
   return buf;
}

void lf_free(void *buf,
	     const char *file, const char *fun, int line)	     
{
   lf_log("F", buf, 0U, file, fun, line);
   free(buf);      
}

char *lf_strdup(const char *buf,
		const char *file, const char *fun, int line)
{
   char *buf2 = strdup(buf);

   lf_log("S", buf2, strlen(buf2) + 1, file, fun, line);
   return buf2;
}
#endif
