#include <stdlib.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <signal.h>
#include <sys/stat.h>
#include <fcntl.h>
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include "lmail.h"
#include "lmail-queue.h"
#include "safeio.h"
#include "trigger.h"
#include "sequence.h"
#include "hacks.h"

static char *dirty_list;

static void ipc_delete(void)
{
   int lazy_sync_semid;
   int dirty_list_shmid;
   
   if ((lazy_sync_semid = semget(LAZY_SYNC_SEMKEY, 1, 0600)) >= 0) {
      semctl(lazy_sync_semid, 0, IPC_RMID, NULL);
   }
   if (dirty_list != NULL) {
      shmdt(dirty_list);
      dirty_list = NULL;
   }
   if ((dirty_list_shmid = shmget(DIRTY_LIST_SHMKEY, QMAIL_MESS_SPLIT, 0600)) >= 0) {
      shmctl(dirty_list_shmid, IPC_RMID, NULL);
   }
   D(puts("IPC removal"));
}

RETSIGTYPE sig(int sg)
{
   D(printf("Signal %d received\n" __ sg));
   signal(sg, SIG_DFL);
   ipc_delete();
   _exit(EXIT_FAILURE);
#if RETSIGTYPE != void
   return 0;
#endif
}

static void sync_sequence(void)
{
   static time_t last_time;
   time_t now = time(NULL);
   
   if (difftime(now, last_time) > (double) LAZY_SYNC_SEQUENCE_DELAY) {
      last_time = now;
      D(puts("sync_sequence"));
#if 0
      write_sequence();
#endif
   }
}

static int init_dirty_list(void)
{
   int dirty_list_shmid;
   
   if ((dirty_list_shmid = shmget(DIRTY_LIST_SHMKEY, QMAIL_MESS_SPLIT, IPC_CREAT | 0600)) < 0) {
      D(perror("shmget"));
      return -1;
   }
   if ((dirty_list = shmat(dirty_list_shmid, NULL, 0)) == NULL) {
      D(perror("shmat"));
      return -1;
   }
   memset(dirty_list, 0, QMAIL_MESS_SPLIT);
   
   return 0;
}

int main(void)
{
   int nbw;
   int lazy_sync_semid;

   if (change_id(LMAIL_QUEUE_USER, LMAIL_QUEUE_GROUP) < 0) {
      puts(_("Sorry, but I need to change my id to ")
	   LMAIL_QUEUE_USER "/" LMAIL_QUEUE_GROUP " (uid/gid).\n"
	   _("Please run me as root and/or check that this virtual user exists."));
      return -1;
   }
   umask(003);         
   atexit(ipc_delete);
   signal(SIGHUP, sig);
   signal(SIGINT, sig);   
   signal(SIGQUIT, sig);
#ifdef SIGILL   
   signal(SIGILL, sig);
#endif
   signal(SIGFPE, sig);            
   signal(SIGKILL, sig);               
   signal(SIGPIPE, sig);
   signal(SIGTERM, sig);   
   signal(SIGUSR1, sig);      
   signal(SIGUSR2, sig);   
   if ((lazy_sync_semid = semget(LAZY_SYNC_SEMKEY, 1, IPC_CREAT | 0600)) < 0) {
      perror("lazy_sync_semid");
      return -1;
   }
   if (init_dirty_list() < 0) {
      perror("init_dirty_list");
      return -1;
   }
   D(puts("lazy_sync_daemon started"));
   for (;;) {
      sleep(LAZY_SYNC_DELAY);
      while ((nbw = semctl(lazy_sync_semid, 0, GETVAL, 0)) > 0) {
	 D(puts("lazy_sync_daemon : flushing metadata"));
	 semctl(lazy_sync_semid, 0, SETVAL, 0);
	 sync_metadata(dirty_list, QMAIL_MESS_SPLIT);
	 sync_sequence();
	 trigger();
      }      
   }
   return 0;
}

