#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <ctype.h>
#include "lmail.h"
#include "bufferedio.h"
#include "lmail-queue.h"
#include "safeio.h"
#include "hacks.h"
#include "strtools.h"
#include "sendfile.h"
#include "varstring.h"
#include "trigger.h"
#include "lmail-send.h"

#ifndef INTD_COMPATIBILITY

static int move_entry_bypass(const unsigned long found_sequence)
{
   char footer[sizeof MESSAGE_FOOTER - 1 + 4];
   int exit_return = 0;
   char *reisertodo_name = NULL;
   size_t reisertodo_name_size;
   char *info_name = NULL;
   size_t info_name_size;
   char *mess_name = NULL;
   size_t mess_name_size;
   char *reisermess_name = NULL;
   size_t reisermess_name_size;
   char *remote_name = NULL;
   size_t remote_name_size;
   int reisertodo_fd = -1;
   int info_fd = -1;
   int remote_fd = -1;
   size_t intd_size;
   off_t intd_off;
   off_t mess_len;
   VarString *envelope_from = NULL;
   size_t uidgid_skipped = (size_t) 0U;
   size_t envelope_to_size;
   
   reisertodo_name_size = (sizeof QUEUE_PATH_REISERTODO - 1) + 1 +
     size_unsigned_int(found_sequence) + 1;
   if ((reisertodo_name = malloc(reisertodo_name_size)) == NULL) {
      D(perror("malloc"));
      return -1;
   }
   snprintf(reisertodo_name, reisertodo_name_size,
	    "%s/%lu", QUEUE_PATH_REISERTODO, found_sequence);
   reisertodo_name[reisertodo_name_size - 1] = 0;
   D(printf("Moving entry [%s]\n" __ reisertodo_name));
   if ((reisertodo_fd = open(reisertodo_name,
			     O_RDWR | O_NDELAY)) < 0) {
      D(perror("reisertodo_name open"));
      goto exit_failure;
   }
   if (lseek(reisertodo_fd, (off_t) -(sizeof MESSAGE_FOOTER - 1 + 4), SEEK_END) == (off_t) -1) {
      D(perror("lseek"));
      goto exit_failure;
   }
   if (safe_read(reisertodo_fd, footer, sizeof footer) <= 0) {
      D(perror("safe_read"));
      goto exit_failure;
   }
   if (memcmp(footer + 4, MESSAGE_FOOTER, sizeof MESSAGE_FOOTER - 1U) != 0) {
      D(printf("[%s] has no valid footer\n" __ reisertodo_name));
      goto exit_failure;
   }
   intd_size = footer[0] << 24 | footer[1] << 16 | footer[2] << 8 | footer[3];
   intd_off = (off_t) -(intd_size + sizeof MESSAGE_FOOTER - 1 + 4 + 2);
   if ((intd_off = lseek(reisertodo_fd, intd_off, SEEK_END)) == (off_t) - 1) {
      D(perror("lseek"));
      goto exit_failure;
   }
   mess_len = intd_off;
   intd_off += (off_t) 2;
      {
	 char c[2];
	 
	 if (safe_read(reisertodo_fd, c, sizeof c) <= 0) {
	    D(perror("reisertodo safe_read"));
	    goto exit_failure;
	 }
	 if (c[0] != 0 || c[1] != 0) {
	    D(printf("Message/Intd delimiter not found : [%c%c]\n" __
	      c[0] __ c[1]));
	    goto exit_failure;
	 }
      }
      {
	 char c;
	 
	 do {
	    if (safe_read(reisertodo_fd, &c, 1U) <= 0) {
	       D(perror("reisertodo safe_read ('F' seek)"));
	       goto exit_failure;
	    }
	    uidgid_skipped++;
	 } while (c != 'F');
	 uidgid_skipped--;
	 if ((envelope_from = var_string_new()) == NULL) {
	    D(puts("envelope_from var_string_new"));
	    goto exit_failure;
	 }
	 if (var_string_cat(envelope_from, "F") < 0) {
	    D(perror("var_string_cat"));
	    goto exit_failure;
	 }
	 for (;;) {
	    char s[2] = { 0, 0 };
	    
	    if (safe_read(reisertodo_fd, s, 1U) <= 0) {
	       D(perror("safe_read"));
	       goto exit_failure;
	    }
	    if (s[0] == 0) {
	       break;
	    }
	    if (var_string_cat(envelope_from, s) < 0) {
	       D(perror("var_string_cat"));
	       goto exit_failure;
	    }
	 }
	 D(printf("Envelope = From [%s]\n", envelope_from->str));
      }
   info_name_size = (sizeof QUEUE_PATH_INFO - 1) + 1 +
     (STRSIZE_QMAIL_MESS_SPLIT - 1) + 1 + size_unsigned_long(found_sequence) + 1;
   if ((info_name = malloc(info_name_size)) == NULL) {
      D(perror("malloc"));
      goto exit_failure;
   }
   snprintf(info_name, info_name_size, "%s/%lu/%lu",
	    QUEUE_PATH_INFO, found_sequence % (unsigned long) QMAIL_MESS_SPLIT,
	    found_sequence);
   info_name[info_name_size - 1] = 0;
   D(printf("Creating info file [%s]\n" __ info_name));
   if ((info_fd = open(info_name, O_WRONLY | O_CREAT | O_EXCL | O_NDELAY,
		       0640)) < 0) {
      D(perror("open"));
      goto exit_failure;
   }
   if (safe_write(info_fd, envelope_from->str, envelope_from->size) < 0) {
      close(info_fd);      
      info_fd = -1;
      unlink(info_name);
      goto exit_failure;
   }
   close(info_fd);
   remote_name_size = (sizeof QUEUE_PATH_REMOTE - 1) + 1 +
     (STRSIZE_QMAIL_MESS_SPLIT - 1) + 1 + size_unsigned_long(found_sequence) + 1;
   if ((remote_name = malloc(remote_name_size)) == NULL) {
      D(perror("malloc"));
      goto exit_failure;
   }
   snprintf(remote_name, remote_name_size, "%s/%lu/%lu",
	    QUEUE_PATH_REMOTE, found_sequence % (unsigned long) QMAIL_MESS_SPLIT,
	    found_sequence);
   remote_name[remote_name_size - 1] = 0;
   D(printf("Creating remote file [%s]\n" __ remote_name));
   envelope_to_size = intd_size - envelope_from->size - uidgid_skipped;
   if (envelope_to_size == (size_t) 0U) {
      D(puts("No recipient ?!"));
      goto exit_failure;
   }
   if ((remote_fd = open(remote_name, O_WRONLY | O_CREAT | O_EXCL | O_NDELAY,
			 0640)) < 0) {
      D(perror("open"));
      goto exit_failure;
   }
      {
	 off_t to_off;
	 
	 to_off = intd_off + envelope_from->size + uidgid_skipped;
	 if (linux_sendfile(remote_fd, reisertodo_fd, &to_off, envelope_to_size) ==
	     (off_t) -1 || fsync(remote_fd) < 0) {
	    D(perror("sendfile/fsync"));
	    close(remote_fd);
	    remote_fd = -1;
	    unlink(remote_name);
	    goto exit_failure;
	 }
      }
   if (ftruncate(reisertodo_fd, mess_len) < 0) {
      D(perror("ftruncate"));
      goto exit_failure;
   }
   mess_name_size = (sizeof QUEUE_PATH_MESS - 1) + 1 +
     (STRSIZE_QMAIL_MESS_SPLIT - 1) + 1 + size_unsigned_long(found_sequence) + 1;
   if ((mess_name = malloc(mess_name_size)) == NULL) {
      D(perror("malloc"));
      goto exit_failure;
   }
   snprintf(mess_name, mess_name_size, "%s/%lu/%lu",
	    QUEUE_PATH_MESS, found_sequence % (unsigned long) QMAIL_MESS_SPLIT,
	    found_sequence);
   mess_name[mess_name_size - 1] = 0;
   D(printf("Moving message to [%s]\n" __ mess_name));   
   if (link(reisertodo_name, mess_name) < 0) {
      D(perror("link"));
      goto exit_failure;
   }
   reisermess_name_size = (sizeof QUEUE_PATH_REISERMESS - 1) + 1 +
     (STRSIZE_QMAIL_REISERMESS_SPLIT - 1) + 1 + size_unsigned_long(found_sequence) + 1;
   if ((reisermess_name = malloc(reisermess_name_size)) == NULL) {
      D(perror("malloc"));
      goto exit_failure;
   }
   snprintf(reisermess_name, reisermess_name_size, "%s/%lu/%lu",
	    QUEUE_PATH_REISERMESS,
	    found_sequence % (unsigned long) QMAIL_REISERMESS_SPLIT,
	    found_sequence);
   reisermess_name[reisermess_name_size - 1] = 0;
   D(printf("Removing [%s]\n" __ reisermess_name));
   if (unlink(reisertodo_name) < 0 ||
       unlink(reisermess_name) < 0) {
      D(perror("unlink"));
      goto exit_failure;
   }
   D(puts("moved"));  
    
   goto exit_success;
   
   exit_failure:
   exit_return = -1;
   
   exit_success:
   if (reisertodo_fd >= 0) {
      close(reisertodo_fd);
   }
   if (info_fd >= 0) {
      close(info_fd);
   }
   if (remote_fd >= 0) {
      close(remote_fd);
   }
   if (reisertodo_name != NULL) {
      free(reisertodo_name);
   }
   if (info_name != NULL) {
      free(info_name);
   }
   if (remote_name != NULL) {
      free(remote_name);
   }
   if (reisermess_name != NULL) {
      free(reisermess_name);
   }
   if (envelope_from != NULL) {
      var_string_free(envelope_from);
   }

   return 0;
}

static int move_entry_compat(const unsigned long found_sequence)
{
   char footer[sizeof MESSAGE_FOOTER - 1 + 4];
   int exit_return = 0;
   char *reisertodo_name = NULL;
   size_t reisertodo_name_size;
   char *intd_name = NULL;
   size_t intd_name_size;
   char *mess_name = NULL;
   size_t mess_name_size;
   char *todo_name = NULL;
   size_t todo_name_size;
   char *reisermess_name = NULL;
   size_t reisermess_name_size;   
   int reisertodo_fd = -1;
   int intd_fd = -1;
   size_t intd_size;
   off_t intd_off;
   off_t mess_len;
   
   reisertodo_name_size = (sizeof QUEUE_PATH_REISERTODO - 1) + 1 +
     size_unsigned_int(found_sequence) + 1;
   if ((reisertodo_name = malloc(reisertodo_name_size)) == NULL) {
      D(perror("malloc"));
      return -1;
   }
   snprintf(reisertodo_name, reisertodo_name_size,
	    "%s/%lu", QUEUE_PATH_REISERTODO, found_sequence);
   reisertodo_name[reisertodo_name_size - 1] = 0;
   D(printf("Moving entry [%s]\n" __ reisertodo_name));
   if ((reisertodo_fd = open(reisertodo_name,
			     O_RDWR | O_NDELAY)) < 0) {
      D(perror("reisertodo_name open"));
      goto exit_failure;
   }
   if (lseek(reisertodo_fd, (off_t) -(sizeof MESSAGE_FOOTER - 1 + 4), SEEK_END) == (off_t) -1) {
      D(perror("lseek"));
      goto exit_failure;
   }
   if (safe_read(reisertodo_fd, footer, sizeof footer) <= 0) {
      D(perror("safe_read"));
      goto exit_failure;
   }
   if (memcmp(footer + 4, MESSAGE_FOOTER, sizeof MESSAGE_FOOTER - 1U) != 0) {
      D(printf("[%s] has no valid footer\n" __ reisertodo_name));
      goto exit_failure;
   }
   intd_size = footer[0] << 24 | footer[1] << 16 | footer[2] << 8 | footer[3];
   intd_off = (off_t) -(intd_size + sizeof MESSAGE_FOOTER - 1 + 4 + 2);
   if ((intd_off = lseek(reisertodo_fd, intd_off, SEEK_END)) == (off_t) - 1) {
      D(perror("lseek"));
      goto exit_failure;
   }
   mess_len = intd_off;
   intd_off += (off_t) 2;
      {
	 char c[2];
	 
	 if (safe_read(reisertodo_fd, c, sizeof c) <= 0) {
	    D(perror("reisertodo safe_read"));
	    goto exit_failure;
	 }
	 if (c[0] != 0 || c[1] != 0) {
	    D(printf("Message/Intd delimiter not found : [%c%c]\n" __
	      c[0] __ c[1]));
	    goto exit_failure;
	 }
      }
   intd_name_size = (sizeof QUEUE_PATH_INTD - 1) + 1 +
     size_unsigned_int(found_sequence) + 1;
   if ((intd_name = malloc(intd_name_size)) == NULL) {
      D(perror("malloc"));
      return -1;
   }
   snprintf(intd_name, intd_name_size, "%s/%lu", 
	    QUEUE_PATH_INTD, found_sequence);
   intd_name[intd_name_size - 1] = 0;
   D(printf("Creating [%s]\n", intd_name));
   if ((intd_fd = open(intd_name, O_WRONLY | O_CREAT | O_EXCL | O_NDELAY,
		       0640)) < 0) {
      D(perror("open"));
      goto exit_failure;
   }
   if (linux_sendfile(intd_fd, reisertodo_fd, &intd_off, intd_size) == 
       (off_t) -1 || fsync(intd_fd) < 0) {
      D(perror("sendfile/fsync"));
      close(intd_fd);
      intd_fd = -1;
      unlink(intd_name);
      goto exit_failure;
   }
   if (ftruncate(reisertodo_fd, mess_len) < 0) {
      D(perror("ftruncate"));
      goto exit_failure;
   }
   mess_name_size = (sizeof QUEUE_PATH_MESS - 1) + 1 +
     (STRSIZE_QMAIL_MESS_SPLIT - 1) + 1 + size_unsigned_long(found_sequence) + 1;
   if ((mess_name = malloc(mess_name_size)) == NULL) {
      D(perror("malloc"));
      goto exit_failure;
   }
   snprintf(mess_name, mess_name_size, "%s/%lu/%lu",
	    QUEUE_PATH_MESS, found_sequence % (unsigned long) QMAIL_MESS_SPLIT,
	    found_sequence);
   mess_name[mess_name_size - 1] = 0;
   D(printf("Moving message to [%s]\n" __ mess_name));   
   if (link(reisertodo_name, mess_name) < 0) {
      D(perror("link"));
      goto exit_failure;
   }
   todo_name_size = (sizeof QUEUE_PATH_TODO - 1) + 1 +
     size_unsigned_long(found_sequence) + 1;
   if ((todo_name = malloc(todo_name_size)) == NULL) {
      D(perror("malloc"));
      goto exit_failure;
   }
   snprintf(todo_name, todo_name_size, "%s/%lu",
	    QUEUE_PATH_TODO, found_sequence);
   if (link(intd_name, todo_name) < 0) {
      D(perror("link"));
      goto exit_failure;
   }
   reisermess_name_size = (sizeof QUEUE_PATH_REISERMESS - 1) + 1 +
     (STRSIZE_QMAIL_REISERMESS_SPLIT - 1) + 1 + size_unsigned_long(found_sequence) + 1;
   if ((reisermess_name = malloc(reisermess_name_size)) == NULL) {
      D(perror("malloc"));
      goto exit_failure;
   }
   snprintf(reisermess_name, reisermess_name_size, "%s/%lu/%lu",
	    QUEUE_PATH_REISERMESS,
	    found_sequence % (unsigned long) QMAIL_REISERMESS_SPLIT,
	    found_sequence);
   reisermess_name[reisermess_name_size - 1] = 0;
   D(printf("Removing [%s]\n" __ reisermess_name));
   if (unlink(reisertodo_name) < 0 ||
       unlink(reisermess_name) < 0) {
      D(perror("unlink"));
      goto exit_failure;
   }
   D(puts("moved"));  
   goto exit_success;
   
   exit_failure:
   exit_return = -1;
   
   exit_success:
   if (reisertodo_fd >= 0) {
      close(reisertodo_fd);
   }
   if (intd_fd >= 0) {
      close(intd_fd);
   }
   if (reisertodo_name != NULL) {
      free(reisertodo_name);
   }
   if (intd_name != NULL) {
      free(intd_name);
   }
   if (mess_name != NULL) {
      free(mess_name);
   }
   if (todo_name != NULL) {
      free(todo_name);
   }
   
   return exit_return;
}

static int proceed(void)
{
   DIR *dir;
   struct dirent *entry;
   unsigned long found_sequence;      
   
   if ((dir = opendir(QUEUE_PATH_REISERTODO)) == NULL) {
      perror("Unable to open " QUEUE_PATH_REISERTODO);
      return -1;
   }
   while ((entry = readdir(dir)) != NULL) {
      if (isdigit((entry->d_name)[0])) {
	 found_sequence = strtoul(entry->d_name, NULL, 10);
# ifdef QMSEND_COMPAT
	 if (move_entry_compat(found_sequence) < 0) {
	    D(perror("move_entry_compat"));
	 }
# else
	 if (move_entry_bypass(found_sequence) < 0) {
	    D(perror("move_entry_bypass"));
	 }	   
# endif
      }
   }
   closedir(dir);
   
   return 0;
}

int main(void)
{
   if (change_id(LMAIL_SEND_USER, LMAIL_SEND_GROUP) < 0) {
      puts(_("Sorry, but I need to change my id to ")
	     LMAIL_SEND_USER "/" LMAIL_SEND_GROUP " (uid/gid).\n"
	   _("Please run me as root and/or check that this virtual user exists."));
      return -1;
   }
   for (;;) {
      proceed();
      trigger_qmail_send();      
      sleep(ZENO_WAIT);
      trigger_wait_lmail_send();      
   }
   return 0;
}
#else
int main(void)
{
   puts(_("This program is useless because you enabled\n"
	  "the old queue format compatibility.\n"));
   return 0;
}
#endif
