enum {
   CMD_RETURN_QUIT = -1,
   CMD_RETURN_UNKNOWN = -2,
   CMD_RETURN_INVALID = -3,
   CMD_RETURN_FAILURE = -4,
   CMD_RETURN_TOO_MANY_HOPS = -5,
};

typedef struct QmailQueue_ {
   char *mess_filename;
   size_t mess_filename_size;
#ifdef INTD_COMPATIBILITY
   char *intd_filename;
   size_t intd_filename_size;
#endif
   unsigned long inode;
   int mess_fd;
#ifdef INTD_COMPATIBILITY
   int intd_fd;   
#endif
} QmailQueue;

typedef struct State_ {
   BufferedReader *client_br;
   BufferedWriter *client_bw;   
   char *helo;
   char *from;   
   List *to;
   VarString *fromto;
   char *data_buf;
   char *data_buf_pnt;   
   size_t data_buf_size;
   size_t data_size;
   int eof_status;
   int header_status;
   const char *header_received_pnt;
   const char *header_delivered_pnt;
   unsigned int hops;
   QmailQueue qmail_queue;
} State;

enum {
   DATA_INITIAL_EOF_STATUS = 1,
   DATA_INITIAL_HEADER_STATUS = 0,
};

#define SMTP_PORT 25
#define DATA_CHUNK_SIZE 16384
#define MAX_HOPS 100U
#ifndef MAX_CLIENTS
# define MAX_CLIENTS 500
#endif
#ifndef BACKLOG
# define BACKLOG (config.max_clients / 2)
#endif
#define BADMAILFROM_SPLIT 27
#define RELAYCLIENTS_SPLIT 27
#define RCPTHOSTS_SPLIT BADMAILFROM_SPLIT
#define MAX_SHUTDOWN_TIME (60U * 5U)

#define CRLF "\r\n"
#ifndef INTD_COMPATIBILITY
# define MESSAGE_FOOTER "RsrSmTp0"
#endif

#define EXTENSION_DATA_SIZE "SIZE"

typedef struct CmdList_ {
   const char *cmd;
   int (*func)(char *arg, State * const state);
} CmdList;

typedef struct ThreadArgs_ {
   int client_fd;
} ThreadArgs;

typedef struct ConfigFiles_ {
   const char *smtpgreeting;
   size_t databytes;
   short smtpdport;
   unsigned int max_hops;
   int max_clients;
   char has_badmailfrom;
   HashedList badmailfrom;
   char open_relay;
   HashedList rcpthosts_strict, rcpthosts_joker;
   CDB morercpthosts;
   char has_relayclients;   
   HashedList relayclients;
} ConfigFiles;
