#include <stdlib.h>
#include <stdio.h>
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include "lmail.h"
#include "varstring.h"
#include "strtools.h"
#include "rfc.h"

char *rfc_date(void)
{
   static const char * const months[] = {
      "Jan", "Feb", "Mar", "Apr", "May", "Jun",
      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
   };
   char *buf;
   size_t buf_size;
   time_t now = time(NULL);
   struct tm *tm;
   long int diff;
   int sign;
   
   if ((tm = localtime(&now)) == NULL) {
      D(perror("localtime"));
      return NULL;
   }
   buf_size = 2 + 1 + 		       /* day of the month */
     3 + 1 + 			       /* month */
     size_int(tm->tm_year + 1900) + 1 +    /* year */
     2 + 1 + 2 + 1 + 2 + 1 +	       /* hour:min:sec */
     1 + 4			       /* timezone */
     + 1;   
   if ((buf = malloc(buf_size)) == NULL) {
      D(perror("malloc"));
      return NULL;
   }
   diff = -timezone / 60L;
   if (diff > 0) {
      sign = '+';
   } else {
      diff = -diff;
      sign = '-';
   }
   snprintf(buf, buf_size, "%d %s %d %02d:%02d:%02d %c%02ld%02ld",
	    tm->tm_mday, months[tm->tm_mon], tm->tm_year + 1900,
	    tm->tm_hour, tm->tm_min, tm->tm_sec,
	    sign, diff / 60L, diff % 60L);
   buf[buf_size - 1] = 0;
   
   return buf;
}

VarString *rfc_received(const char *remotehost, const char *helo,
			const char *remoteinfo, const char *remoteip,
			const char *localhost, const char *protocol)
{
   VarString *buf;
   char *date;
   
   if ((buf = var_string_new()) == NULL) {
      return NULL;
   }
   if (remotehost == NULL) {
      remotehost = _("unknown");
   }
   if (helo == NULL) {
      helo = _("nobody");
   }
   if (remoteinfo == NULL) {
      remoteinfo = _("unknown");
   }
   if (remoteip == NULL) {
      remoteip = _("nowhere");
   }
   if (localhost == NULL) {
      localhost = _("localhost");
   }
   if (protocol == NULL) {
      protocol = _("unknown protocol");
   }
   if ((date = rfc_date()) == NULL ||
       var_string_cat(buf, "Received: ") != 0 ||       
       var_string_cat(buf, "from ") != 0 ||
       var_string_cat(buf, remotehost) != 0 ||
       var_string_cat(buf, " (HELO ") != 0 ||
       var_string_cat(buf, helo) != 0 ||
       var_string_cat(buf, ") (") != 0 ||
       var_string_cat(buf, remoteinfo) != 0 ||
       var_string_cat(buf, "@") != 0 ||
       var_string_cat(buf, remoteip) != 0 ||
       var_string_cat(buf, ") by ") != 0 ||
       var_string_cat(buf, localhost) != 0 ||
       var_string_cat(buf, " (" PACKAGE "-" VERSION ") ") != 0 ||
       var_string_cat(buf, "with ") != 0 ||
       var_string_cat(buf, protocol) != 0 ||
       var_string_cat(buf, "; ") != 0 ||
       var_string_cat(buf, date) != 0 ||
       var_string_cat(buf, "\n") != 0) {
      var_string_free(buf);
      if (date != NULL) {
	 free(date);
      }
      return NULL;
   }
   free(date);
   
   return buf;
}
	       
