#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "lmail.h"
#include "safeio.h"

#ifndef O_NDELAY
# define O_NDELAY O_NONBLOCK
#endif

#define RETRIES 2

extern int errno;

ssize_t safe_write(const int fd, const void * const buf_, size_t size)
{
   ssize_t res;
   int retry = RETRIES;
   unsigned char *buf = (unsigned char *) buf_;
   
   while (size > 0) {
      if ((res = write(fd, buf, size)) < (ssize_t) 0) {	 
	 if (errno == EINTR) {
	    continue;
	 } else if (errno == EAGAIN) {
	    anew:
	    if (retry == 0) {
	       int z = fcntl(fd, F_GETFL);
	       z &= ~O_NDELAY;
	       fcntl(fd, F_SETFL, z);
	    } else {
	       sleep(1);
	       retry--;
	    }
	    continue;
	 } else {
	    return SAFE_WRITE_ERROR;
	 }
      } else if (res == 0) {
	 goto anew;
      }
      size -= res;
      buf += res;
   }
   return (ssize_t) (buf - (unsigned char *) buf_);
}

ssize_t safe_write_line(const int fd, const char * const buf)
{   
   return safe_write(fd, buf, strlen(buf));
}

ssize_t safe_read(const int fd, void * const buf_, size_t size)
{
   ssize_t res;
   ssize_t written = (ssize_t) 0;
   int retry = RETRIES;      
   unsigned char *buf = (unsigned char *) buf_;

   return read(fd, buf, size);
   while (size > 0) {
      if ((res = read(fd, buf, size)) < (ssize_t) 0) {	 
	 if (errno == EINTR) {
	    continue;
	 } else if (errno == EAGAIN) {
	    if (retry == 0) {
	       int z = fcntl(fd, F_GETFL);
	       z &= ~O_NDELAY;
	       fcntl(fd, F_SETFL, z);
	    } else {
	       sleep(1);
	       retry--;
	    } 
	 } else {
	    return SAFE_READ_ERROR;
	 }
      } else if (res == 0) {
	 return SAFE_READ_EOF;
      } else {
	 return res + written;
      }
      size -= res;
      buf += res;
      written += res;
   }
   return written;
}

ssize_t safe_get_char(const int fd, char * const c)
{
   return safe_read(fd, c, (size_t) 1U);
}

int safe_get_line(const int fd, char *buf, size_t size)
{
   char c;

   if (size < (size_t) 2U) {
      return SAFE_READ_ERROR;
   }
   for (;;) {
      size--;
      if (size == (size_t) 0U) {
	 *buf = 0;
	 return SAFE_READ_OVERFLOW;
      }
      if (safe_get_char(fd, &c) < 0) {
	 *buf = 0;
	 return SAFE_READ_ERROR;
      }
      *buf = c;
      buf++;
      if (c == '\n') {
	 break;
      }
   }
   *buf = 0;
   
   return 0;
}

int file_lock(const int fd, const int type)
{
   struct flock lock;
   
   lock.l_type = type;
   lock.l_start = (off_t) 0;
   lock.l_len = (off_t) 0;
   lock.l_pid = getpid();
   for (;;) {
      if (fcntl(fd, F_SETLKW, &lock) < 0) {
	 if (errno == EINTR) {
	    continue;	    
	 }
      }
      break;
   }   
   
   return 0;
}
