#define DEFINE_SEQUENCE_GLOBALS
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <stdio.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <dirent.h>
#include <ctype.h>
#include "lmail.h"
#include "strtools.h"
#include "safeio.h"
#include "sequence.h"
#include "lmail-queue.h"

extern int total_clients;

static unsigned long
  find_sequence_scandir(const char *path,
			unsigned long highest_sequence)
{
   DIR *dir;
   struct dirent *entry;
   unsigned long found_sequence;   

   if ((dir = opendir(path)) == NULL) {
      return highest_sequence;
   }
   while ((entry = readdir(dir)) != NULL) {
      if (isdigit((entry->d_name)[0])) {
	 found_sequence = strtoul(entry->d_name, NULL, 10);
	 if (found_sequence > highest_sequence) {
	    highest_sequence = found_sequence;
	 }
      }
   }
   closedir(dir);
   
   return highest_sequence;
}

int init_sequence(void)
{
   char *filename;
   size_t filename_size;
   unsigned long highest_sequence;
   int t;   
   struct stat st;

   if (stat(QUEUE_PATH_MESS, &st) < 0 || !S_ISDIR(st.st_mode)) {
      return -1;
   }
   filename_size = (sizeof QUEUE_PATH_MESS - 1U) + 1U +
     (STRSIZE_QMAIL_MESS_SPLIT - 1U) + 1U;
   if ((filename = alloca(filename_size)) == NULL) {
      D(perror("alloca"));
      return -1;
   }
   t = QMAIL_MESS_SPLIT;
   highest_sequence = 0UL;
   do {
      t--;
      snprintf(filename, filename_size, "%s/%d", QUEUE_PATH_MESS, t);
      filename[filename_size - 1] = 0;
      highest_sequence = find_sequence_scandir(filename, highest_sequence);
   } while (t > 0);
   ALLOCA_FREE(filename);
   
#ifndef INTD_COMPATIBILITY
   if (stat(QUEUE_PATH_REISERMESS, &st) < 0 || !S_ISDIR(st.st_mode)) {
      return -1;
   }
   filename_size = (sizeof QUEUE_PATH_REISERMESS - 1U) + 1U +
     (STRSIZE_QMAIL_REISERMESS_SPLIT - 1U) + 1U;
   if ((filename = alloca(filename_size)) == NULL) {
      D(perror("alloca"));
      return -1;
   }
   t = QMAIL_REISERMESS_SPLIT;
   do {
      t--;
      snprintf(filename, filename_size, "%s/%d", QUEUE_PATH_REISERMESS, t);
      filename[filename_size - 1] = 0;
      highest_sequence = find_sequence_scandir(filename, highest_sequence);
   } while (t > 0);
   ALLOCA_FREE(filename); 
#endif
   
   highest_sequence = find_sequence_scandir(QUEUE_PATH_TODO, highest_sequence);   
   highest_sequence = find_sequence_scandir(QUEUE_PATH_INTD, highest_sequence);
#ifndef INTD_COMPATIBILITY
   highest_sequence = find_sequence_scandir(QUEUE_PATH_REISERTODO, highest_sequence);
#endif
   highest_sequence = highest_sequence + 1UL;
   D(printf("Start sequence : [%lu]\n" __ highest_sequence));
   global_sequence = highest_sequence;
   
   return 0;
}

#ifdef PREALLOCATE
void *preallocate(void *thread_args)
{
   static char preallocate_buf[PREALLOCATE_SIZE];
   unsigned long sequence = global_sequence + PREALLOCATE_GAP;
   size_t mess_filename_size;
   char *mess_filename;
   register char *mess_filename_strpnt;
   int fd;
   int sleeze = 0;
   unsigned int counter = PREALLOCATE_CHUNK;
   
   (void) thread_args;
   if (global_preallocate_sequence > sequence) {
      if (sequence > PREALLOCATE_CHUNK) {
	 return NULL;
      }
   }
#ifdef INTD_COMPATIBILITY   
   mess_filename_size =
     (sizeof QUEUE_PATH_MESS - 1U) + 1U /* / */ + 
     (STRSIZE_QMAIL_MESS_SPLIT - 1U) + 1U /* / */ + STRSIZE_ULONG + 1U;
#else
   mess_filename_size =
     (sizeof QUEUE_PATH_REISERMESS - 1U) + 1U /* / */ + 
     (STRSIZE_QMAIL_REISERMESS_SPLIT - 1U) + 1U /* / */ + STRSIZE_ULONG + 1U;
#endif
   if ((mess_filename = alloca(mess_filename_size)) == NULL) {
      return NULL;
   }   
   D(puts("Starting preallocation"));
   while (total_clients < 2 && counter > 0U) {
      STR2_INIT(mess_filename, mess_filename_strpnt);
#ifdef INTD_COMPATIBILITY   
      STR2_UNSAFE_STRCAT_L(mess_filename_strpnt, QUEUE_PATH_MESS "/", sizeof (QUEUE_PATH_MESS "/"));
      STR2_UNSAFE_ULINTCAT(mess_filename_strpnt, sequence % (unsigned long) QMAIL_MESS_SPLIT);      
#else
      STR2_UNSAFE_STRCAT_L(mess_filename_strpnt, QUEUE_PATH_REISERMESS "/", sizeof (QUEUE_PATH_REISERMESS "/"));
      STR2_UNSAFE_ULINTCAT(mess_filename_strpnt, sequence % (unsigned long) QMAIL_REISERMESS_SPLIT);
#endif

      STR2_UNSAFE_CHARCAT(mess_filename_strpnt, '/');
      STR2_UNSAFE_ULINTCAT(mess_filename_strpnt, sequence);
      STR2_UNSAFE_FINALIZE(mess_filename_strpnt);
      if ((fd = open(mess_filename, O_WRONLY | O_CREAT | O_EXCL | 
		     O_NDELAY, 0660)) < 0) {
	 if (sleeze == 0) {
	    sleeze++;
	    continue;
	 } else {
	    break;
	 }
      }
      safe_write(fd, preallocate_buf, sizeof preallocate_buf);
      close(fd);
      sequence++;
      counter--;
   }
   global_preallocate_sequence = sequence;
   D(printf("Preallocated until [%s] clients = [%d]\n", mess_filename,
	    total_clients));
   ALLOCA_FREE(mess_filename);
   
   return NULL;
}
#endif
