#include <stdlib.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <stdio.h>
#include "lmail.h"
#include "sync-queue.h"

static int lock;

static void getlock(void)
{
   while (lock != 0) {
      D(puts("Waiting for the lock"));
   }
   lock = 1;
}

static void unlock(void)
{
   lock = 0;
}

int sync_queue_init(SyncQueue * const sq)
{
   if (sq == NULL) {
      D(puts("Null sync queue"));
      return -1;
   }
   sq->first = sq->last = NULL;
   
   return 0;
}

int sync_queue_push(SyncQueue * const sq, const int fd1
#ifdef INTD_COMPATIBILITY
		    ,const int fd2
#endif
		    )
{
   SyncNode *new_node;
   
   getlock();
#ifdef INTD_COMPATIBILITY   
   if (sq == NULL || fd1 < 0 || fd2 < 0) {
      D(printf("Pushing odd things to a sync queue : fd1 [%d] fd2 [%d]\n" __
	       fd1 __ fd2));
      unlock();
      return -1;
   }
#else
   if (sq == NULL || fd1 < 0) {
      D(printf("Pushing odd things to a sync queue : fd1 [%d]\n" __
	       fd1));
      unlock();
      return -1;
   }   
#endif
   if ((new_node = malloc(sizeof *new_node)) == NULL) {
      perror("malloc");
      unlock();      
      return -1;
   }
   new_node->next = NULL;
   new_node->fd1 = fd1;
#ifdef INTD_COMPATIBILITY
   new_node->fd2 = fd2;
#endif
   if (sq->last == NULL) {
      sq->first = sq->last = new_node;
   } else {
      sq->last->next = new_node;
      sq->last = new_node;
   }
   unlock();
   
   return 0;
}

int sync_queue_pop(SyncQueue * const sq, SyncNode * const sync_node)
{
   SyncNode *sn;
   
   getlock();
   if (sq == NULL || sync_node == NULL) {
      D(puts("Null arguments to sync_queue_pop"));
      unlock();
      return -1;
   }
   if ((sn = sq->first) == NULL) {
      unlock();      
      return 1;
   }
   *sync_node = *sn;
   sq->first = sn->next;
   if (sq->first == NULL) {
      sq->last = NULL;
   }
   free(sn);   
   unlock();
   
   return 0;
}

int sync_queue_delete(SyncQueue * const sq)
{
   SyncNode foo;
   
   if (sq == NULL) {
      D(puts("Null sync queue to delete"));
      return -1;
   }
   while (sync_queue_pop(sq, &foo) == 0) {
   }

   return 0;
}
