#include <stdlib.h>
#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <errno.h>
#include "lmail-queue.h"
#include "lmail.h"
#include "safeio.h"
#include "trigger.h"

extern int errno;

int sync_dir(const char *dir)
{
   int fd;

   if ((fd = open(dir, O_RDONLY | O_NDELAY)) < 0) {
      return -1;
   }
   fsync(fd);
   close(fd);
   
   return 0;
}

int sync_metadata(char *dirty_list, size_t dirty_list_size)
{
#ifndef NO_METADATA_SYNC   
   char path_mess[(sizeof QUEUE_PATH_MESS) + 1 + STRSIZE_QMAIL_MESS_SPLIT];
   size_t z = (size_t) 0U;
   
   do {
      if (dirty_list[z] != 0) {
	 snprintf(path_mess, sizeof path_mess, "%s/%d", QUEUE_PATH_MESS, z);
	 D(printf("Syncing dirty mess dir : %s\n" __ path_mess));
	 path_mess[sizeof path_mess - 1] = 0;
	 if (sync_dir(path_mess) < 0) {
	    return -1;
	 }
	 dirty_list[z] = 0;
      }
      z++;
   } while (z < dirty_list_size);
   if ((sync_dir(QUEUE_PATH_INTD) + 
       sync_dir(QUEUE_PATH_TODO)) != 0) {  
      return -1;
   }
#else
   (void) dirty_list;
   (void) dirty_list_size;
#endif
   
   return 0;
}

void trigger_qmail_send(void)
{
   int fd;
   char c = 0;
   
   if ((fd = open(QUEUE_PATH_TRIGGER_QMAIL_SEND, O_WRONLY | O_NDELAY)) >= 0) {
      write(fd, &c, 1U);
      close(fd);
   }
}

void trigger_lmail_send(void)
{
   int fd;
   char c = 0;
   
   if ((fd = open(QUEUE_PATH_TRIGGER_LMAIL_SEND, O_WRONLY | O_NDELAY)) >= 0) {
      write(fd, &c, 1U);
      close(fd);
   }
}

void trigger_wait_lmail_send(void)
{
   static int fd = -1;
   static fd_set readfds;
   static struct timeval timeout;
   
   if (fd >= 0) {
      close(fd);
   }
   if ((fd = open(QUEUE_PATH_TRIGGER_LMAIL_SEND, O_RDONLY)) < 0) {
      perror(_("Unable to open " QUEUE_PATH_TRIGGER_LMAIL_SEND));
      _exit(EXIT_FAILURE);
   }
   FD_ZERO(&readfds);
   FD_SET(fd, &readfds);
   timeout.tv_sec = TRIGGER_SELECT_TIMEOUT;
   timeout.tv_usec = 0;
   for (;;) {
      if (select(fd + 1, &readfds, NULL, NULL, &timeout) < 0) {
	 if (errno == EINTR) {
	    continue;
	 }
	 D(perror("select"));
	 return;
      }
      break;
   }
   D(puts("Triggered"));
   close(fd);
   fd = -1;
}
