#include <stdlib.h>
#include <string.h>
#include "lmail.h"
#include "varstring.h"

VarString *var_string_new(void)
{
   VarString *vs;
   
   if ((vs = malloc(sizeof *vs)) == NULL) {
      return NULL;
   }
   vs->str = NULL;
   vs->size = vs->mem_size = 0U;
   
   return vs;
}

void var_string_free(VarString * const vs)
{
   if (vs != NULL) {
      if (vs->str != NULL) {
	 free(vs->str);
      }
      free(vs);
   }
}

int var_string_cat(VarString * const vs, const char *s)
{
   size_t ss;
   
   if (vs == NULL || s == NULL) {
      return VAR_STRING_FAILURE;
   }
   ss = strlen(s) + 1U;
   if (vs->mem_size <= ss) {
      vs->mem_size += VAR_STRING_CHUNK_SIZE;
      if ((vs->str = realloc(vs->str, vs->mem_size)) == NULL) {
	 vs->mem_size -= VAR_STRING_CHUNK_SIZE;
	 return VAR_STRING_FAILURE;
      }
   }
   if (vs->size == 0U) {
      memcpy(vs->str, s, ss);
      vs->size += ss;      
   } else {
      memcpy(vs->str + vs->size - 1, s, ss);
      vs->size += (ss - 1);            
   } 
   
   return 0;
}

int var_string_binadd(VarString * const vs, const char *s, size_t ss)
{
   if (vs == NULL || s == NULL) {
      return VAR_STRING_FAILURE;
   }
   if (ss == 0) {
      return 0;
   }
   if (vs->mem_size <= ss) {
      vs->mem_size += VAR_STRING_CHUNK_SIZE;
      if ((vs->str = realloc(vs->str, vs->mem_size)) == NULL) {
	 vs->mem_size -= VAR_STRING_CHUNK_SIZE;
	 return VAR_STRING_FAILURE;
      }
   }
   memcpy(vs->str + vs->size, s, ss);
   vs->size += ss;      
   
   return 0;
}

int var_string_add(VarString * const vs, const char *s)
{
   return var_string_binadd(vs, s, strlen(s) + 1);
}

int var_string_compress(VarString * const vs)
{
   if (vs == NULL) {
      return VAR_STRING_FAILURE;
   }
   if (vs->size > (size_t) 0U && vs->size != vs->mem_size) {
      if ((vs->str = realloc(vs->str, vs->size)) == NULL) {
	 return VAR_STRING_FAILURE;
      }
      vs->mem_size = vs->size;
   }
   return 0;
}
